﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2011 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.SymbolicExpressionTreeEncoding;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Problems.DataAnalysis.Symbolic.TimeSeriesPrognosis {
  [Item("Mean squared error Evaluator", "Calculates the mean squared error of a symbolic time-series prognosis solution.")]
  [StorableClass]
  public class SymbolicTimeSeriesPrognosisSingleObjectiveMeanSquaredErrorEvaluator : SymbolicTimeSeriesPrognosisSingleObjectiveEvaluator {
    [StorableConstructor]
    protected SymbolicTimeSeriesPrognosisSingleObjectiveMeanSquaredErrorEvaluator(bool deserializing) : base(deserializing) { }
    protected SymbolicTimeSeriesPrognosisSingleObjectiveMeanSquaredErrorEvaluator(SymbolicTimeSeriesPrognosisSingleObjectiveMeanSquaredErrorEvaluator original, Cloner cloner)
      : base(original, cloner) {
    }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new SymbolicTimeSeriesPrognosisSingleObjectiveMeanSquaredErrorEvaluator(this, cloner);
    }

    public SymbolicTimeSeriesPrognosisSingleObjectiveMeanSquaredErrorEvaluator() : base() { }

    public override bool Maximization { get { return false; } }

    public override IOperation Apply() {
      var solution = SymbolicExpressionTreeParameter.ActualValue;
      IEnumerable<int> rows = GenerateRowsToEvaluate();

      double quality = Calculate(SymbolicDataAnalysisTreeInterpreterParameter.ActualValue,
        solution,
        EstimationLimitsParameter.ActualValue.Lower, EstimationLimitsParameter.ActualValue.Upper,
        ProblemDataParameter.ActualValue,
        rows, HorizonParameter.ActualValue.Value);
      QualityParameter.ActualValue = new DoubleValue(quality);

      return base.Apply();
    }

    public static double Calculate(ISymbolicTimeSeriesPrognosisExpressionTreeInterpreter interpreter, ISymbolicExpressionTree solution, double lowerEstimationLimit, double upperEstimationLimit, ITimeSeriesPrognosisProblemData problemData, IEnumerable<int> rows, int horizon) {
      double[] alpha;
      double[] beta;
      DetermineScalingFactors(solution, problemData, interpreter, rows, out alpha, out beta);
      var scaledSolution = Scale(solution, alpha, beta);

      var meanSquaredErrorCalculators = Enumerable.Range(0, problemData.TargetVariables.Count())
        .Select(i => new OnlineMeanSquaredErrorCalculator()).ToArray();

      var allPredictedContinuations = interpreter.GetSymbolicExpressionTreeValues(scaledSolution, problemData.Dataset,
                                                                                  problemData.TargetVariables.ToArray(),
                                                                                  rows, horizon);
      var allContinuationsEnumerator = allPredictedContinuations.GetEnumerator();
      var rowsEnumerator = rows.GetEnumerator();
      while (allContinuationsEnumerator.MoveNext() & rowsEnumerator.MoveNext()) {
        var componentEnumerator = allContinuationsEnumerator.Current.GetEnumerator();
        var mseCalculatorsEnumerator = meanSquaredErrorCalculators.AsEnumerable().GetEnumerator();
        var targetVarEnumerator = problemData.TargetVariables.GetEnumerator();
        while (componentEnumerator.MoveNext() & mseCalculatorsEnumerator.MoveNext() & targetVarEnumerator.MoveNext()) {
          var stepEnumerator = componentEnumerator.Current.GetEnumerator();
          var actualContinuationEnumerator = problemData.Dataset.GetDoubleValues(targetVarEnumerator.Current,
                                                                       Enumerable.Range(rowsEnumerator.Current, horizon)).GetEnumerator();
          while (stepEnumerator.MoveNext() & actualContinuationEnumerator.MoveNext()) {
            double e = Math.Min(upperEstimationLimit, Math.Max(lowerEstimationLimit, stepEnumerator.Current));
            mseCalculatorsEnumerator.Current.Add(actualContinuationEnumerator.Current, e);
            if (mseCalculatorsEnumerator.Current.ErrorState == OnlineCalculatorError.InvalidValueAdded)
              return double.NaN;
          }
        }
      }
      var meanCalculator = new OnlineMeanAndVarianceCalculator();
      foreach (var calc in meanSquaredErrorCalculators) {
        if (calc.ErrorState != OnlineCalculatorError.None) return double.NaN;
        meanCalculator.Add(calc.MeanSquaredError);
      }
      //int i = 0;
      //foreach (var targetVariable in problemData.TargetVariables) {
      //  var predictedContinuations = allPredictedContinuations.Select(v => v.ElementAt(i));
      //  for (int h = 0; h < horizon; h++) {
      //    OnlineCalculatorError errorState;
      //    meanCalculator.Add(OnlineMeanSquaredErrorCalculator.Calculate(predictedContinuations
      //                                                                    .Select(x => x.ElementAt(h))
      //                                                                    .LimitToRange(lowerEstimationLimit,
      //                                                                                  upperEstimationLimit),
      //                                                                  actualContinuations.Select(x => x.ElementAt(h)),
      //                                                                  out errorState));
      //    if (errorState != OnlineCalculatorError.None) return double.NaN;
      //  }
      //}
      return meanCalculator.MeanErrorState == OnlineCalculatorError.None ? meanCalculator.Mean : double.NaN;
    }

    private static ISymbolicExpressionTree Scale(ISymbolicExpressionTree solution, double[] alpha, double[] beta) {
      var clone = (ISymbolicExpressionTree)solution.Clone();
      int n = alpha.Length;
      for (int i = 0; i < n; i++) {
        var parent = clone.Root.GetSubtree(0);
        var rpb = clone.Root.GetSubtree(0).GetSubtree(i);
        var scaledRpb = MakeSum(
          MakeProduct(rpb,
            MakeConstant(beta[i], clone.Root.Grammar), clone.Root.Grammar),
            MakeConstant(alpha[i], clone.Root.Grammar), clone.Root.Grammar);
        parent.RemoveSubtree(i);
        parent.InsertSubtree(i, scaledRpb);
      }
      return clone;
    }

    private static ISymbolicExpressionTreeNode MakeSum(ISymbolicExpressionTreeNode a, ISymbolicExpressionTreeNode b, ISymbolicExpressionTreeGrammar grammar) {
      var sum = grammar.Symbols.Where(s => s is Addition).First().CreateTreeNode();
      sum.AddSubtree(a);
      sum.AddSubtree(b);
      return sum;
    }

    private static ISymbolicExpressionTreeNode MakeProduct(ISymbolicExpressionTreeNode a, ISymbolicExpressionTreeNode b, ISymbolicExpressionTreeGrammar grammar) {
      var prod = grammar.Symbols.Where(s => s is Multiplication).First().CreateTreeNode();
      prod.AddSubtree(a);
      prod.AddSubtree(b);
      return prod;
    }

    private static ISymbolicExpressionTreeNode MakeConstant(double c, ISymbolicExpressionTreeGrammar grammar) {
      var node = (ConstantTreeNode)grammar.Symbols.Where(s => s is Constant).First().CreateTreeNode();
      node.Value = c;
      return node;
    }

    private static void DetermineScalingFactors(ISymbolicExpressionTree solution, ITimeSeriesPrognosisProblemData problemData, ISymbolicTimeSeriesPrognosisExpressionTreeInterpreter interpreter, IEnumerable<int> rows, out double[] alpha, out double[] beta) {
      string[] targetVariables = problemData.TargetVariables.ToArray();
      int nComponents = targetVariables.Length;
      alpha = new double[nComponents];
      beta = new double[nComponents];
      var oneStepPredictions = interpreter.GetSymbolicExpressionTreeValues(solution, problemData.Dataset,
                                                                     targetVariables, rows,
                                                                     1);
      for (int i = 0; i < nComponents; i++) {
        var target = problemData.Dataset.GetDoubleValues(targetVariables[i], rows);
        var prediction = oneStepPredictions.Select(v => v.ElementAt(i).First());
        OnlineCalculatorError errorState;
        OnlineLinearScalingParameterCalculator.Calculate(prediction, target, out alpha[i], out beta[i], out errorState);
        if (errorState != OnlineCalculatorError.None) {
          alpha[i] = 0;
          beta[i] = 1;
        }
      }
    }

    public override double Evaluate(IExecutionContext context, ISymbolicExpressionTree tree, ITimeSeriesPrognosisProblemData problemData, IEnumerable<int> rows) {
      SymbolicDataAnalysisTreeInterpreterParameter.ExecutionContext = context;
      EstimationLimitsParameter.ExecutionContext = context;
      HorizonParameter.ExecutionContext = context;

      double mse = Calculate(SymbolicDataAnalysisTreeInterpreterParameter.ActualValue, tree, EstimationLimitsParameter.ActualValue.Lower, EstimationLimitsParameter.ActualValue.Upper, problemData, rows, HorizonParameter.ActualValue.Value);

      HorizonParameter.ExecutionContext = null;
      SymbolicDataAnalysisTreeInterpreterParameter.ExecutionContext = null;
      EstimationLimitsParameter.ExecutionContext = null;

      return mse;
    }
  }
}
